"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from atomic_reactor.plugin import PostBuildPlugin
from atomic_reactor.rpm_util import rpm_qf_args, parse_rpm_output
from docker.errors import APIError


__all__ = ('PostBuildRPMqaPlugin', )


class PostBuildRPMqaPlugin(PostBuildPlugin):
    key = "all_rpm_packages"
    is_allowed_to_fail = False
    sep = ';'

    def __init__(self, tasker, workflow, image_id, ignore_autogenerated_gpg_keys=True):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        """
        # call parent constructor
        super(PostBuildRPMqaPlugin, self).__init__(tasker, workflow)
        self.image_id = image_id
        self.ignore_autogenerated_gpg_keys = ignore_autogenerated_gpg_keys

        self._container_ids = []

    def run(self):
        # If another component has already filled in the image component list, skip
        if self.workflow.image_components is not None:
            return None

        plugin_output = self.gather_output()

        # gpg-pubkey are autogenerated packages by rpm when you import a gpg key
        # these are of course not signed, let's ignore those by default
        if self.ignore_autogenerated_gpg_keys:
            self.log.debug("ignore rpms 'gpg-pubkey'")
            plugin_output = [x for x in plugin_output if not x.startswith("gpg-pubkey" + self.sep)]

        volumes = []
        for container_id in self._container_ids:
            volumes.extend(self.tasker.get_volumes_for_container(container_id))

            try:
                self.tasker.remove_container(container_id)
            except APIError:
                self.log.warning("error removing container %s (ignored):", container_id,
                                 exc_info=True)

        for volume_name in volumes:
            try:
                self.tasker.remove_volume(volume_name)
            except APIError:
                self.log.warning("error removing volume %s (ignored):", volume_name, exc_info=True)

        self.workflow.image_components = parse_rpm_output(plugin_output)

        return plugin_output

    def gather_output(self):
        for _ in range(5):
            container_id = self.tasker.run(
                self.image_id,
                command=rpm_qf_args(),
                create_kwargs={"entrypoint": "/bin/rpm", "user": "root"},
                start_kwargs={},
            )
            self._container_ids.append(container_id)
            self.tasker.wait(container_id)
            output = self.tasker.logs(container_id, stream=False)

            if output:
                return output

        raise RuntimeError('Unable to gather list of installed packages in container')
