"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.


Pre build plugin which injects custom yum repository in dockerfile.
"""
import os
from atomic_reactor.constants import YUM_REPOS_DIR, RELATIVE_REPOS_PATH, INSPECT_CONFIG
from atomic_reactor.plugin import PreBuildPlugin
from atomic_reactor.util import df_parser


def add_yum_repos_to_dockerfile(yumrepos, df, inherited_user):
    if df.baseimage is None:
        raise RuntimeError("No FROM line in Dockerfile")

    # Determine the USER the final image should end with
    final_user_line = "USER " + inherited_user if inherited_user else None
    # Look for the last USER after the last FROM... by looking in reverse
    for insndesc in reversed(df.structure):
        if insndesc['instruction'] == 'USER':
            final_user_line = insndesc['content']  # we will reuse the line verbatim
            break
        if insndesc['instruction'] == 'FROM':
            break  # no USER specified in final stage

    # Insert the ADD line at the beginning of each stage
    df.add_lines(
        "ADD %s* '%s'" % (RELATIVE_REPOS_PATH, YUM_REPOS_DIR),
        all_stages=True, at_start=True
    )

    # Insert line(s) to remove the repos
    cleanup_lines = [
        "RUN rm -f " +
        " ".join(["'%s'" % repo for repo in yumrepos])
    ]
    # If needed, change to root in order to RUN rm, then change back.
    if final_user_line:
        cleanup_lines.insert(0, "USER root")
        cleanup_lines.append(final_user_line)

    df.add_lines(*cleanup_lines)


class InjectYumRepoPlugin(PreBuildPlugin):
    key = "inject_yum_repo"
    is_allowed_to_fail = False

    def __init__(self, tasker, workflow):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        """
        # call parent constructor
        super(InjectYumRepoPlugin, self).__init__(tasker, workflow)
        self.host_repos_path = os.path.join(self.workflow.builder.df_dir, RELATIVE_REPOS_PATH)

    def run(self):
        """
        run the plugin
        """
        yum_repos = {k: v for k, v in self.workflow.files.items() if k.startswith(YUM_REPOS_DIR)}
        if not yum_repos:
            return
        # absolute path in containers -> relative path within context
        repos_host_cont_mapping = {}
        host_repos_path = os.path.join(self.workflow.builder.df_dir, RELATIVE_REPOS_PATH)
        self.log.info("creating directory for yum repos: %s", host_repos_path)
        os.mkdir(host_repos_path)

        for repo, repo_content in self.workflow.files.items():
            repo_basename = os.path.basename(repo)
            repo_relative_path = os.path.join(RELATIVE_REPOS_PATH, repo_basename)
            repo_host_path = os.path.join(host_repos_path, repo_basename)
            self.log.info("writing repo to '%s'", repo_host_path)
            with open(repo_host_path, "wb") as fp:
                fp.write(repo_content.encode("utf-8"))
            self.log.debug("%s\n%s", repo, repo_content.strip())
            repos_host_cont_mapping[repo] = repo_relative_path

        # Find out the USER inherited from the base image
        inspect = self.workflow.builder.inspect_base_image()
        inherited_user = inspect[INSPECT_CONFIG].get('User', '')
        df = df_parser(self.workflow.builder.df_path, workflow=self.workflow)
        add_yum_repos_to_dockerfile(repos_host_cont_mapping, df, inherited_user)
